// SPDX-FileCopyrightText: Copyright (c) 2025 Xronos Inc.
// SPDX-License-Identifier: BSD-3-Clause

#ifndef XRONOS_TELEMETRY_ATTRIBUTE_MANAGER_HH
#define XRONOS_TELEMETRY_ATTRIBUTE_MANAGER_HH

#include <cstdint>
#include <mutex>
#include <optional>
#include <string>
#include <unordered_map>
#include <variant>

#include "runtime/reactor_element.hh"

namespace xronos::telemetry {

/// A variant type that lists all possible attribute types.
// The order of types matters for the pybind11 bindings. See
// https://pybind11.readthedocs.io/en/stable/advanced/cast/stl.html#c-17-library-containers
using AttributeValue = std::variant<std::string, bool, std::int64_t, double>;

using AttributeMap = std::unordered_map<std::string, AttributeValue>;

class AttributeManager {

private:
  std::unordered_map<std::uint64_t, AttributeMap> attribute_maps_{};
  mutable std::mutex mutex_{};

  auto try_emplace_and_get_attribute_map(const reactor::ReactorElement& element) -> AttributeMap&;

public:
  /** Set a single attribute for the given element. */
  void set_attribute(const reactor::ReactorElement& element, std::string_view key, AttributeValue value);
  /** Set multiple attributes for the given element. */
  void set_attributes(const reactor::ReactorElement& element, AttributeMap attributes);

  /** Add a new attribute for the given element.
   *
   * Similar to `set_attribute`, but only adds new attributes that are not previously present.
   * Returns `true` if the attribute was inserted.
   */
  auto add_attribute(const reactor::ReactorElement& element, std::string_view key, const AttributeValue& value) -> bool;

  /** Add multiple attributes for the given element.
   *
   * Similar to `set_attributes`, but only adds new attributes that are not previously present.
   * Returns `true` if all attributes were inserted.
   */
  auto add_attributes(const reactor::ReactorElement& element, AttributeMap attributes) -> bool;

  /** Retrieve the attributes of the given element. */
  auto get_attributes(const reactor::ReactorElement& element) const -> std::optional<AttributeMap>;

  //  * This also includes all attributes of the element's containers. If a
  //  * attribute key is set for an element and one of its containers, then
  //  * the element's attribute value takes precedence.
  //  */
  // auto get_attributes(const reactor::ReactorElement& element) const -> AttributeMap;

  template <class MapType, class ValueType>
  auto
  get_attributes_converted(const reactor::ReactorElement& element,
                           std::function<ValueType(const AttributeValue&)> convert) const -> std::optional<MapType>;
};

template <class MapType, class ValueType>
auto AttributeManager::get_attributes_converted(const reactor::ReactorElement& element,
                                                std::function<ValueType(const AttributeValue&)> convert) const
    -> std::optional<MapType> {
  std::lock_guard<std::mutex> lock{mutex_};
  auto iterator = attribute_maps_.find(element.uid());
  if (iterator != attribute_maps_.end()) {
    MapType attributes{};
    for (const auto& [key, value] : iterator->second) {
      attributes.try_emplace(key, convert(value));
    }
    return attributes;
  }
  return std::nullopt;
}

}; // namespace xronos::telemetry

#endif // XRONOS_TELEMETRY_ATTRIBUTE_MANAGER_HH
